/* **************************************************************************

  SPACE.C, a library of general spatial and geometric utilities.
  Space.c contains jasort, an N-dimensional spatial sorting algorithm
of my own invention.

Copyright (C) Feb 19, 1999 John Conner Atkeson

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

For further information, contact:
John Atkeson jcatkeson@juno.com
210 Drummond Place, Apt 19, Norfolk Va. 23507 USA

************************************************************************** */

/*
  Jasort is essentially a hybrid between QuickSort and an N-dimensional
binary quadtree.  The "tree" it produces however is perfectly balanced
from birth, and is a straight array of xyz data without need for any
kind of pointers to hold it together.
  So it is very efficient with both speed and memory.

  This implimentation sorts an index to the data for flexibility,
rather than sorting the xyz data itself.  But you can alter the code
to do away with the indexing if you like.

*/

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include "space.h"

/* to swap values into ascending order. */
#define sort2(a,b,low,high) if (a<b){low=a;high=b;}else{low=b;high=a;}

bool InSpace(XYZType Point,CubeT *ThisBox,byte Dimensions)
/* **************************************************************** 
* Determine if a point is inside a sorted n-dimensional cube. 
**************************************************************** */
{
  byte Cnt;
  bool Inside;
  
  Inside=true;Cnt=0;
  while((Cnt<Dimensions)&&Inside){
    if((Point.Loc[Cnt]<ThisBox->L[0][Cnt])||(ThisBox->L[1][Cnt]<Point.Loc[Cnt])){
      Inside=false;
    }
    Cnt++;
  }
  return(Inside);
}/* InSpace */

double Hypot(double Xdif,double Ydif)
/* *************************************************************
* Returns the hypotenuse between the two leg lengths.          *
************************************************************* */
{
  return(sqrt((Xdif *Xdif)+(Ydif *Ydif)));
}

double AngleOf(double Xv,double Yv)
/**************************************************************/
/* Returns angle formed by run (Xv) and rise (Yv).            */
/**************************************************************/
{
  double Angle;
  if(Xv<0){
    Angle=atan(Yv/Xv)+3.14159265360;/* 180  3.141592653589793238462643383279; */
  }else{
    if(Xv==0){/*Wall is vertical. */
      if(Yv>0){
        Angle=1.57079632680;/* 90 Vertical up. */
      }else{
        Angle=4.71238898040;/* 270 Vertical down. */
      }
    }else{
      if(Yv<0){
        Angle=atan(Yv/Xv)+6.28318530720;/* 360 */
      }else{
        Angle=atan(Yv/Xv);
      }
    }
  }
  return(Angle);
}/* AngleOf */

double AngleSpread(double X1,double Y1,double XVert,double YVert,double X2,double Y2)
/**************************************************************/
/* Gives the angle made by three points.                      */
/**************************************************************/
{
  double Ang1,Ang2;
  Ang1=AngleOf(X1-XVert,Y1-YVert);Ang2=AngleOf(X2-XVert,Y2-YVert);
  if(Ang1>Ang2)Ang1=Ang1-TwoPi;
  return(Ang2-Ang1);
}

void SetCubeSz(CubeT *Z)
/************************************************************/
/* Set the box's Wdt,Hgt,Dep to match its limits.           */
/************************************************************/
{
  int Dim;
  for(Dim=0;Dim<MaxDims;Dim++){
    Z->Sz[Dim]=Z->L[1][Dim]-Z->L[0][Dim];
  }
}/* SetCubeSz */
void SetCube(double x1,double y1,double z1,double x2,double y2,double z2,CubeT *Z)
/************************************************************/
/* Set the limits on a box.                                 */
/************************************************************/
{
  Z->L[0][0]=x1;Z->L[0][1]=y1;Z->L[0][2]=z1;
  Z->L[1][0]=x2;Z->L[1][1]=y2;Z->L[1][2]=z2;
  SetCubeSz(Z);
}
void MoveCube(double movex,double movey,double movez,CubeT *Z)
/************************************************************
*  Translate (move) a cube by a relative vector.
************************************************************/
{
  int cnt;
  for (cnt=0;cnt<2;cnt++){/* from min corner to max corner of cube */
    Z->L[cnt][0]+=movex;Z->L[cnt][1]+=movey;Z->L[cnt][2]+=movez;
  }
  SetCubeSz(Z);
}/* MoveCube */
void ExpandCube(double factor, CubeT *Z)
/************************************************************
*  Multiply a cube's size by some factor; stays centered.
************************************************************/
{
  int dimcnt;
  double center,halfsz;
  for (dimcnt=0;dimcnt<MaxDims;dimcnt++){
    center=(Z->L[1][dimcnt]+Z->L[0][dimcnt])/2.0;
    halfsz=Z->Sz[dimcnt]*factor/2.0;
    Z->L[0][dimcnt]= center-halfsz;
    Z->L[1][dimcnt]= center+halfsz;
  }
  SetCubeSz(Z);
}/* ExpandCube */
void BorderCube(CubeT *Z,double border)
/**************************************************************
*  Add a border to every face of a cube.
**************************************************************/
{
  int cnt;
  for (cnt=0;cnt<MaxDims;cnt++){
    Z->L[0][cnt]-=border; Z->L[1][cnt]+=border;
  }
  SetCubeSz(Z);
}/* BorderCube */

void SortCube(CubeT *Cube)
/*********************************************************************/
/* Sorts a cube so that the second corner is greater than the first. */
/*********************************************************************/
{
  unsigned int Cnt;double Temp;
  for(Cnt=0;Cnt<MaxDims;Cnt++){
    if(Cube->L[0][Cnt]>Cube->L[1][Cnt]){
      Temp=Cube->L[0][Cnt];
      Cube->L[0][Cnt]=Cube->L[1][Cnt];
      Cube->L[1][Cnt]=Temp;
    }
    Cube->Sz[Cnt]=Cube->L[1][Cnt]-Cube->L[0][Cnt];
  }
}

void SetBoxWH(CubeT *bx)
/************************************************************/
/* Set the box's Wdt,Hgt to match its limits.               */
/************************************************************/
{
  bx->Sz[0]=bx->L[1][0]-bx->L[0][0];bx->Sz[1]=bx->L[1][1]-bx->L[0][1];
  /*    bx->Sz[2]=bx->L[1][2]-bx->L[0][2];  // z stuff */
}
void SetBox(double x1,double y1,double x2,double y2,CubeT *bx)
/************************************************************/
/* Set the limits on a box.                                 */
/************************************************************/
{
  /*    SetCube(x1,y1,0,x1,y1,0,bx); */
  bx->L[0][0]=x1;bx->L[0][1]=y1;bx->L[1][0]=x2;bx->L[1][1]=y2;
  /*    bx->L[0][2]=0; bx->L[1][2]=0; */
  SetBoxWH(bx);
}
void OrthoFitBox(BoxType ABox,BoxType *BBox)
/* *************************************************************** */
/* * Fit all of BBox tightly inside ABox without distorting it.  * */
/* * Similar to SHOW command in Rocky Mountain Basic, except that* */
/* * it only works on the box data structure.  You have to set   * */
/* * the screen windows yourself with the results.               * */
/* *************************************************************** */
{
  double axtoy,bxtoy;
  bxtoy=BBox->Sz[0]/BBox->Sz[1];axtoy=ABox.Sz[0]/ABox.Sz[1];
  if(fabs(bxtoy)>fabs(axtoy)){/* If b is wider and squatter than a. */
    BBox->L[1][1]=BBox->L[0][1]+BBox->Sz[0] *fabs((ABox.Sz[1]/ABox.Sz[0]));
  }else{/* Else if b is taller and narrower than a. */
    BBox->L[1][0]=BBox->L[0][0]+BBox->Sz[1] *fabs((ABox.Sz[0]/ABox.Sz[1]));
  }
  SetBoxWH(BBox);
}/* OrthoFitBox */

void FlipCubeDim(CubeT *bx,int dimension)
/* ********************************************************************
*  N-dimensional; flip the polarity of one dimension of the cube.
******************************************************************** */
{
  double temp;
  temp=bx->L[0][dimension]; 
  bx->L[0][dimension]=bx->L[1][dimension];
  bx->L[1][dimension]=temp;
  SetBoxWH(bx);
}/* FlipCubeDim */

void PolarizeND(const CubeT *ABox,CubeT *BBox,int dimension)
/* ********************************************************************
*   N-dimensional; set the *direction* of BBox's proportions to 
* equal the direction of ABox's proportions.  E.G. if ABox has
* negative width, then set BBox to have negative width; if ABox
* has positive height, then set BBox to also have positive height.
*   BBox's actual size and location are not changed.
******************************************************************** */
{ 
  double sign,swap; int DimCnt;
  for(DimCnt=0;DimCnt<dimension;DimCnt++){
    sign=ABox->Sz[DimCnt]*BBox->Sz[DimCnt];
    if (sign<0){/* if signs are different, invert BBox limits. */
      swap=BBox->L[0][DimCnt]; BBox->L[0][DimCnt]=BBox->L[1][DimCnt]; BBox->L[1][DimCnt]=swap; 
    }
  }/* end for. */
  SetCubeSz(BBox);
}/* PolarizeND */

void OrthoFitND(const CubeT *ABox,CubeT *BBox,int dimension)
/* ********************************************************************
*   N-dimensional; set BBox's proportions to equal ABox's proportions, 
* while keeping all of old BBox's contents tightly inside the new BBox.
******************************************************************** */
{ 
  double MinRat,Center,HalfSz,Ratio[MaxDims];int DimCnt;
  MinRat=maxint32;/* absurd to lose first comparison. */
  for(DimCnt=0;DimCnt<dimension;DimCnt++){/* first get ratios of sizes, and narrowest dimension. */
    Ratio[DimCnt]=fabs(ABox->Sz[DimCnt]/BBox->Sz[DimCnt]);
    if(Ratio[DimCnt]<MinRat){MinRat=Ratio[DimCnt];}
  }/* end for. */
  for(DimCnt=0;DimCnt<dimension;DimCnt++){
    BBox->Sz[DimCnt]=(BBox->Sz[DimCnt]) *(Ratio[DimCnt]/MinRat);
    Center=(BBox->L[1][DimCnt]+BBox->L[0][DimCnt])/2.0;
    HalfSz=BBox->Sz[DimCnt]/2;
    BBox->L[0][DimCnt]=Center-HalfSz;BBox->L[1][DimCnt]=Center+HalfSz;
  }/* end for. */
  SetCubeSz(BBox);
}/* OrthoFitND */

void OrthoFitCube(const CubeT *ABox,CubeT *BBox)
/* ********************************************************************
*   Set BBox's proportions to equal ABox's proportions, while keeping 
* all of old BBox's contents tightly inside the new BBox.
******************************************************************** */
{
  double MinRat,Center,HalfSz,Ratio[MaxDims];int DimCnt;
  MinRat=maxint32;/* absurd to lose first comparison. */
  for(DimCnt=0;DimCnt<MaxDims;DimCnt++){/* first get ratios of sizes, and narrowest dimension. */
    Ratio[DimCnt]=fabs(ABox->Sz[DimCnt]/BBox->Sz[DimCnt]);
    if(Ratio[DimCnt]<MinRat){MinRat=Ratio[DimCnt];}
  }/* end for. */
  for(DimCnt=0;DimCnt<MaxDims;DimCnt++){
    BBox->Sz[DimCnt]=(BBox->Sz[DimCnt]) *(Ratio[DimCnt]/MinRat);
    Center=(BBox->L[1][DimCnt]+BBox->L[0][DimCnt])/2.0;
    HalfSz=BBox->Sz[DimCnt]/2;
    BBox->L[0][DimCnt]=Center-HalfSz;BBox->L[1][DimCnt]=Center+HalfSz;
  }/* end for. */
  SetCubeSz(BBox);
}/* OrthoFitCube */

void AreaFit2d(CubeT *Z,double newWdivH)
/**************************************************************
*   Set a box to new proportions, but keep the old
* surface area.
**************************************************************/
{
  double area,newwdt,newhgt;

  area=Z->Sz[0] * Z->Sz[1];/* oldhgt*oldwdt; */
  newhgt=sqrt(area * newWdivH); /* newWdivH=proportions, littlewdt/littlehgt */
  if (newhgt==0) newhgt=fudge;/* workaround divide-by-zero. */
  newwdt=area/newhgt;
  Z->L[1][0]=Z->L[0][0]+newwdt; Z->L[1][1]=Z->L[0][1]+newhgt;
  SetCubeSz(Z);
}/* AreaFit2d */

void Prj(double xloc,double yloc,const FrameT *F,double *Outx,double *Outy)
/**************************************************************/
/* Maps point from abstract coords to screen.                 */
/**************************************************************/
{
#if 1
  ( *Outx)=((xloc-F->Wn.L[0][0])/F->Wn.Sz[0]) *F->Vw.Sz[0]+F->Vw.L[0][0];
  ( *Outy)=((yloc-F->Wn.L[0][1])/F->Wn.Sz[1]) *F->Vw.Sz[1]+F->Vw.L[0][1];
#else
  ( *Outx)=(xloc*F->scale[0])+F->ShfScaleShf[0]; ( *Outy)=(yloc*F->scale[1])+F->ShfScaleShf[1];
#endif
}

void DePrj(double xloc,double yloc,const FrameT *F,double *Inx,double *Iny)/* Descale from window.   */
/**************************************************************/
/* Maps point from screen to abstract coords.                 */
/**************************************************************/
{
  ( *Inx)=((xloc-F->Vw.L[0][0])/F->Vw.Sz[0]) *F->Wn.Sz[0]+F->Wn.L[0][0];
  ( *Iny)=((yloc-F->Vw.L[0][1])/F->Vw.Sz[1]) *F->Wn.Sz[1]+F->Wn.L[0][1];
}

void ResetFrame(FrameT *Fr)
/* ************************************************************
*  Update frame's scale values based on its Wn and Vw boxes.
************************************************************ */
{
  int cnt;
  double temp;
  for (cnt=0;cnt<MaxDims;cnt++){
    temp=Fr->Wn.Sz[cnt];  
    if (fabs(temp)<=0.0000001) temp=fudge;/* fudge to catch divide-by-zero. */
    Fr->scale[cnt]=Fr->Vw.Sz[cnt]/temp;
    Fr->ShfScaleShf[cnt] = (Fr->scale[cnt]*(-Fr->Wn.L[0][cnt]))+Fr->Vw.L[0][cnt];    
  }
}/* ResetFrame */

void SetFrame(CubeT *Wn,CubeT *Vw,FrameT *Fr)
/* ************************************************************
************************************************************ */
{
  int cnt;
  double temp;
  memcpy(&(Fr->Wn),Wn,sizeof(CubeT)); memcpy(&(Fr->Vw),Vw,sizeof(CubeT));
  for (cnt=0;cnt<MaxDims;cnt++){
    temp=Wn->Sz[cnt];
    if (fabs(temp)<=0.0000001) temp=fudge;/* fudge to catch divide-by-zero. */
    Fr->scale[cnt]=Vw->Sz[cnt]/temp;
    Fr->ShfScaleShf[cnt] = (Fr->scale[cnt]*(-Wn->L[0][cnt]))+Vw->L[0][cnt];    
  }
}/* SetFrame */

void ReproportionFrame(double x0,double y0,double z0,
  double x1,double y1,double z1,FrameT *Fr)
/**************************************************************
*   Reset frame's viewport side and reproportion window side
* to match, preserving all scale and aspect ratio.
**************************************************************/
{ 
  int cnt; 
  SetCube(x0,y0,z0,x1,y1,z1,&(Fr->Vw));
  for (cnt=0;cnt<MaxDims;cnt++){
    Fr->Wn.L[1][cnt]=Fr->Wn.L[0][cnt] + (Fr->Vw.Sz[cnt]/Fr->scale[cnt]);
    printf("frame size:%lf, frame scale:%lf:\n",Fr->Vw.Sz[cnt],Fr->scale[cnt]);
  }
  SetCubeSz(&(Fr->Wn));
}/* ReproportionFrame */

void Rotate(double OldX,double OldY,double Angle,double *NewX,double *NewY)
/* ************************************************************
* Rotate X and Y around 0,0 by Angle radians.
************************************************************ */
{
  double SinRot,CosRot;
  SinRot=sin(Angle);CosRot=cos(Angle);
  *NewX=OldX *CosRot-OldY *SinRot; *NewY=OldY *CosRot+OldX *SinRot;
}

void SetLimits(PListT *AllSurv,CubeT *MinMax)
/* ***********************************************************
* get spatial bounds of data and set them in MinMax.
*********************************************************** */
{
  intdex Cnt,NumPoints;byte DimCnt;
  /* set MinMax bounds to ridiculous inverted extremes. */
  NumPoints=AllSurv->NumPoints;
  if (NumPoints>0){
    for(DimCnt=0;DimCnt<MaxDims;DimCnt++){
      MinMax->L[0][DimCnt]=maxint32; MinMax->L[1][DimCnt]=-maxint32;
    }
    for(Cnt=0;Cnt<NumPoints;Cnt++){
      for(DimCnt=0;DimCnt<MaxDims;DimCnt++){
        /* This is important, before sorting your data test its min, max bounds */
        /* and keep them in MinMax[0] and MinMax[1]. */
        if(AllSurv->P[Cnt].Loc[DimCnt]<MinMax->L[0][DimCnt]){
          MinMax->L[0][DimCnt]=AllSurv->P[Cnt].Loc[DimCnt];
        }
        if(AllSurv->P[Cnt].Loc[DimCnt]>MinMax->L[1][DimCnt]){
          MinMax->L[1][DimCnt]=AllSurv->P[Cnt].Loc[DimCnt];
        }
      }
    }
  }else{/* if no data, set MinMax to mostly zero. */
    for(DimCnt=0;DimCnt<MaxDims;DimCnt++){
      MinMax->L[0][DimCnt]=0; MinMax->L[1][DimCnt]=fudge;
    }
  }
  SetCubeSz(MinMax);/* set MinMax's size to match its bounds. */
}/* end of SetLimits. */

void plist_blankrefs(PListT *points)
/* ***********************************************************
*  Set all of the ref fields in a plist to NULL.
*********************************************************** */
{ 
  intdex num_points;
  XYZType *P;
  intdex pcnt;
  
  num_points=points->NumPoints;
  P=points->P;
  
  for (pcnt=0;pcnt<num_points;pcnt++){
    P[pcnt].ref=NULL;
  }
}/* end of plist_blankrefs */
void InitPList(PListT *Space,intdex NumElements)
/* ***********************************************************
*  Create a point list to size, all zeros.
*********************************************************** */
{ 
  Space->P=calloc(NumElements,sizeof(XYZType));  
  if(Space->P==NULL){
    printf(" Totally out of memory for index! \r\n");
    printf("Press cntrl-c to avoid a computer crash. \r\n");
    getchar();
  }
  Space->NumPoints=NumElements;
}/* end of InitPList */
void FreePList(PListT *Space)
/* ***********************************************************
*  Dispose of a PList to free memory.
*********************************************************** */
{
  free(Space->P);
  Space->P=NULL; Space->NumPoints=0;
}/* FreePList */

void MakeBlankSortVect(VectT *Vec,intdex NumElements)
/* ***********************************************************
* Make a big blank sortvector, no contents for quick creaton.
*********************************************************** */
{
  intdex Cnt;
  Vec->V=(VRay *)calloc(NumElements,sizeof(intdex));
  
  if(Vec->V==NULL){
    printf(" Totally out of memory for index! \r\n");
    printf("Press cntrl-c to avoid a computer crash. \r\n");
    getchar();
  }
  memset(Vec->V,0,NumElements*sizeof(intdex));
  Vec->NumPoints=0;
}/* MakeBlankSortVect */

void InitSortVect(VectT *Vec,intdex NumElements)
/* ***********************************************************
* Initialize a sort index (vector) to 0,1,2,3... NumElements-1.
*********************************************************** */
{
  intdex Cnt;
  Vec->V=(VRay *)calloc(NumElements,sizeof(intdex));
  
  if(Vec->V==NULL){
    printf(" Totally out of memory for index! \r\n");
    printf("Press cntrl-c to avoid a computer crash. \r\n");
    getchar();
  }
  {VRay *V=Vec->V;/* with - do */
    for(Cnt=0;Cnt<NumElements;Cnt++){
      V[Cnt]=Cnt;
    }
  }/* end with */
  Vec->NumPoints=NumElements;
}/* end of InitSortVect */
void FreeSortVect(VectT *Vec)
/* ***********************************************************
*********************************************************** */
{
  free(Vec->V);
  Vec->V=NULL; Vec->NumPoints=0;
}/* FreeSortVect */

void FlipSortVect(VectT *Vec)
/* ***********************************************************
*  Flip all the elements at the beginning of the array with the end.
*********************************************************** */
{
  intdex halfway=Vec->NumPoints>>1;
  intdex *V=&(Vec->V[0]);
  intdex cnt,topcnt,temp;
  
  topcnt=Vec->NumPoints-1;
  for (cnt=0;cnt<halfway;cnt++){
    temp=V[cnt]; V[cnt]=V[topcnt]; V[topcnt]=temp; /* swap top and bottom */
    topcnt--;
  }
}/* end of FlipSortVect */


static void SwapIt(VRay *V,intdex FirstLoc,intdex SecondLoc)
/************************************************************
* Swap two sortvector records (long ints).                 *
 ************************************************************/
{
  intdex Temp;
  Temp=V[FirstLoc];V[FirstLoc]=V[SecondLoc];V[SecondLoc]=Temp;
}

static double KeyOf(SortPairT *sp,intdex Index,byte Dimension)
/* ************************************************************ */
/* Return the sort key of our data, given the sort index and */
/* "key index", IE the dimension it is sorted by. */
/* ************************************************************ */
{
  intdex SubjIndex;/* SubjIndex is the actual data array subscript. */
  SubjIndex=sp->SVect->V[Index];
  return(sp->PList->P[SubjIndex].Loc[Dimension]);
}

static void Partition(SortPairT *sp,intdex Low,intdex High,intdex *PivotLoc,byte Dimension)
/* *************************************************************
* Sorts a section of the point array into < or > pivotval.     *
************************************************************* */
{
  double PivotVal;intdex DataIndex,LastSmall;/* ulong SubjectIndex; */

  SwapIt(sp->SVect->V,Low,(Low+High+1)>>1);/* shift right 1 is same as div 2 but faster. */
  PivotVal=KeyOf(sp,Low,Dimension);
  LastSmall=Low;
  for(DataIndex=Low+1;DataIndex<=High;DataIndex++){
    if(KeyOf(sp,DataIndex,Dimension)<PivotVal){
      LastSmall++;
      SwapIt(sp->SVect->V,LastSmall,DataIndex);
    }
  }
  SwapIt(sp->SVect->V,Low,LastSmall);
  *PivotLoc=LastSmall;
}/* Partition */

static void MedioPartition(SortPairT *sp,intdex MLow,intdex MHigh,intdex *MedianLoc,byte Dimension)
/* ***********************************************************
* Partition vector right down the population middle, or      *
* statistical median.  Used for spatial plane sort.          *
*********************************************************** */
{
  intdex PivotLoc=-1;
  *MedianLoc=(MLow+MHigh)>>1;/* shift right 1 is same as div 2 but faster. */
  while(PivotLoc!=*MedianLoc){
    Partition(sp,MLow,MHigh,&PivotLoc,Dimension);
    if( *MedianLoc<PivotLoc){/* Partition first half, on next iteration. */
      MHigh=PivotLoc-1;
    }else{/* Partition second half, on next iteration. */
      MLow=PivotLoc+1;
    }
  }
  /* Now the median is found, and we are partitioned on the median. */
}/* MedioPartition */

static void jasort_recurse(SortPairT *sp,intdex Low,intdex High,byte SortLevel)
/* ***********************************************************
*  Recursive part, most closely resembles quicksort here,
* except for balanced partitioning.
*********************************************************** */
{
  intdex MedianLoc;
  if(Low<High){
    MedioPartition(sp,Low,High,&MedianLoc,SortLevel%SortDims);
    jasort_recurse(sp,Low,MedianLoc-1,SortLevel+1);
    jasort_recurse(sp,MedianLoc+1,High,SortLevel+1);
  }
}/* jasort_recurse */

void jasort(PListT *AllSurv2,VectT *XYSort2)
/* ***********************************************************
* Sorts multidimensionally.
*********************************************************** */
{
  SortPairT sp;

  sp.PList=AllSurv2; sp.SVect=XYSort2;

  SetLimits(AllSurv2,&MinMax);
  jasort_recurse(&sp,0,AllSurv2->NumPoints-1,0);
}/* End of jasort. */

void Treeverse(VectT *SVect,PListT *PList,CubeT *Zone,intdex MinLoc,intdex MaxLoc,size_t Level,VectT *outray)
/* **************************************************************
* Recursive traverse of a "SortDims"-dimensional sorted region.
* All points found are returned indexed in outray.
*  Use TraverseZone as front end to this.
*************************************************************** */
{
  XYZType *P;
  intdex MedLoc,SubjIndex,LevelMod=Level%SortDims;

  MedLoc=(MinLoc+MaxLoc)>>1;/* shift right 1 is same as div 2 but faster. */
  
  /* get data index from sort pointer.*/
  SubjIndex=SVect->V[MedLoc];/* median point is sample. */
  P=&(PList->P[SubjIndex]);/* with-do for sample point. */
  
  if((P->Loc[LevelMod]>=Zone->L[0][LevelMod])&&(MinLoc<MedLoc)){
    Treeverse(SVect,PList,Zone,MinLoc,MedLoc-1,Level+1,outray);/* Explore lower half. */
  }
  if(InSpace(*P,Zone,SortDims)){/* if in zone, add to output list. */
    outray->V[outray->NumPoints]=SubjIndex;
    outray->NumPoints++;
  }
  if((P->Loc[LevelMod]<=Zone->L[1][LevelMod])&&(MedLoc<MaxLoc)){
    Treeverse(SVect,PList,Zone,MedLoc+1,MaxLoc,Level+1,outray);/* Explore upper half. */
  }
}/* End of Treeverse. */
void TraverseZone(VectT *SVect,PListT *PList,CubeT *Zone,VectT *outray)
/* **************************************************************
*  Non-recursive front end to Treeverse.
* All points found are returned indexed in outray.
*************************************************************** */
{
  intdex NumPoints=SVect->NumPoints;
  if (NumPoints>0){
    Treeverse(SVect,PList,Zone,0,NumPoints-1,0,outray);
  }else{
    outray->NumPoints=0;
  }
}/* End of TraverseZone. */

#if 0

/*  begin Tin routines. */
 move to tin.c

bool NoFilter(XYZType *Pt){/* default empty filter. */
  return(true);
}
int GetNeighbors(PListT *AllSurv2,VectT *XYSort2,intdex CenterDex,LogicType *Filter,VectT *Nbrs)
/* ******************************************************************* */
/* Returns all the TIN-neighbors around a particular sounding point. */
/*  Point is passed as an array index, and neighbors are returned */
/*  in an array of array indices. */
/* ******************************************************************* */
{
  contextT context;
  intdex ThisNbrDex,FirstDex;
  intdex NbrCnt=0;/* neighbor counter. */
  bool Err=0;/* if neighborhood is incomplete     */
  VectT paramv;

  MakeBlankSortVect(&paramv,XYSort2->NumPoints);

  if(Filter==NULL){
    Filter=&NoFilter;
  }
  context.points=AllSurv2; context.sort=XYSort2; context.paramv=&paramv;

  FindClosest(&context,CenterDex,&ThisNbrDex,Filter);/* pythagorean closest. */
  FirstDex=ThisNbrDex;/* save first neighbor to detect finished loop. */
  
  while(1){/* repeat... until */
    /* add it to list and increment counter. */
    Nbrs->V[NbrCnt]=ThisNbrDex;NbrCnt++;
    
    FindThird(&context,CenterDex,ThisNbrDex,&ThisNbrDex,Filter);/* triangle search. */
    
    /* break out of loop if next neighbor is the first one, */
    /* we have completed the loop. */
    if(ThisNbrDex==FirstDex){
      Err=0;break;/* looped neighborhood is complete, exit. */
    }else{
      if(ThisNbrDex<=-1){
        Err=1;break;/* no neighbors or edge found, leave. */
      }
    }
  }/* end while. */
  Nbrs->NumPoints=NbrCnt;
  FreeSortVect(&paramv);
  return(Err);
}/* end of GetNeighbors. */
#endif
/* end tin routines. */

/* Begin Polygon support. */
void create_blank_poly(polyT *p,int numofsides)
/* **************************************************************
*  Allocate memory to a brand new (uninitialized) polygon.
************************************************************** */
{
  p->vtex=calloc(numofsides,sizeof(xyT)); p->ncorners=0; p->closed=false;
}/* create_blank_poly */
void realloc_poly(polyT *p,int numofsides)
/* **************************************************************
*  Reallocate memory to increase potential size of polygon.
************************************************************** */
{
  p->vtex=realloc(p->vtex,numofsides*sizeof(xyT));
}/* realloc_poly */
void freepoly(polyT *p)
/* **************************************************************
*  dispose of a polygon to free memory.
************************************************************** */
{
  if (p->vtex!=NULL){
    free(p->vtex); p->vtex=NULL;
  }
  p->ncorners=0; p->closed=false;
}/* freepoly */
void poly_addvertex(polyT *p,double newx,double newy)
/* **************************************************************
*  Add a vertex to a polygon.
************************************************************** */
{
  /* do some kind of range checking and realloc first. */
  p->vtex[p->ncorners].xy[0]=newx; p->vtex[p->ncorners].xy[1]=newy;
  p->ncorners++;
}/* poly_addvertex */
void RegPoly(double xorg,double yorg,double radius,int numofsides,polyT *p)
/* **************************************************************
* Make a regular polygon.
************************************************************** */
{
  double angle;
  double nowx,nowy;
  int linenum,vcnt;

  create_blank_poly(p,numofsides);
  angle = TwoPi/(double)numofsides;/* angle of increment. */
  linenum = 0;

  for (vcnt=0;vcnt<numofsides;vcnt++){
    nowx = cos(angle*linenum)*radius; nowy = sin(angle*linenum)*radius;
    p->vtex[vcnt].xy[0] = nowx+xorg; p->vtex[vcnt].xy[1] = nowy+yorg;
    linenum++;
  }
  p->ncorners = numofsides; p->closed=true;
}/* RegPoly */

void get_poly_limits(polyT *p,CubeT *minmax)
/* **************************************************************
*  Get min-max bounding box that encloses polygon.
************************************************************** */
{
  static int dimension=2;
  intdex vcnt,numofsides,dimcnt;

  memset(minmax,0,sizeof(CubeT));
  /* set minmax bounds to ridiculous inverted extremes. */
  for(dimcnt=0;dimcnt<dimension;dimcnt++){
    minmax->L[0][dimcnt]=maxint32; minmax->L[1][dimcnt]=-maxint32;
  }
  numofsides=p->ncorners;
  for (vcnt=0;vcnt<numofsides;vcnt++){
    for (dimcnt=0;dimcnt<dimension;dimcnt++){
      /* look for minimum. */
      if (minmax->L[0][dimcnt]>p->vtex[vcnt].xy[dimcnt]){
        minmax->L[0][dimcnt]=p->vtex[vcnt].xy[dimcnt];
      }
      /* look for maximum. */
      if (minmax->L[1][dimcnt]<p->vtex[vcnt].xy[dimcnt]){
        minmax->L[1][dimcnt]=p->vtex[vcnt].xy[dimcnt];
      }
    }
  }
  SetCubeSz(minmax);
}/* get_poly_limits */

#if 0

bool inside_poly(polyT *poly,double xloc,double yloc)
/* **************************************************************
*    Must be a closed polygon.
*    Test if xloc, yloc is inside or outside of a polygon.
*  We draw an imaginary line from our point downward to infinity.
*  If the line crosses an odd number of polygon walls, the point
*  must be inside.  Otherwise it is outside.
************************************************************** */
{
  int pntcnt,ncorners;
  double test;
  bool Inside,Done;
  double x1,y1,x2,y2;
  ncorners=poly->ncorners;

/* test=0 if point is on line, test>0 if above, test<0 if below.*/
  Done = false; Inside = false;

  x1 = poly->vtex[ncorners-1].xy[0]; y1 = poly->vtex[ncorners-1].xy[1];
  pntcnt = 0;
  while ((!Done) && (pntcnt<ncorners)){
    test = -1.0;

    x2 = poly->vtex[pntcnt].xy[0]; y2 = poly->vtex[pntcnt].xy[1];
    if (x1=x2){/* if line is vertical*/
      if ((y1<yloc && yloc<=y2) || (y2<=yloc && yloc<y1)){
        test = xloc - x1;
      }
    }else{
      if ((x1<xloc && xloc<=x2) || (x2<=xloc && xloc<x1)){/* if bracketed by x1 and x2*/
        test = (yloc-y1) - ( (xloc-x1) * (y2-y1)/(x2-x1) );
      }
    }
    if(test>0){/* we crossed a line, so flip flag */
      Inside = !Inside;
    }else{
      if (test == 0){/* right on line, so officially inside. */
        Inside = true; Done = true;
      }
    }
    x1 = x2; y1 = y2;
    pntcnt++;
  } /* end while */
  return(Inside);
}/* inside_poly */

#else

bool inside_poly(polyT *poly,double xloc,double yloc)
/* **************************************************************
*    Must be a closed polygon.
*    Test if xloc, yloc is inside or outside of a polygon.
*  We draw an imaginary line from our point downward to infinity.
*  If the line crosses an odd number of polygon walls, the point
*  must be inside.  Otherwise it is outside.
************************************************************** */
{
  /* flag for whether we've crossed an odd number of line segments to reach xloc,yloc. */
  int inside=false;
  double oldx,oldy,newx,newy,x1,y1,x2,y2;/* various line endpoints. */
  double slope,base,intersectY;/* line formula and intersect values. */
  int ncorners,lcnt;

  ncorners=poly->ncorners;
  oldx=poly->vtex[ncorners-1].xy[0]; oldy=poly->vtex[ncorners-1].xy[1];
  for (lcnt=0;lcnt<ncorners;lcnt++){
    newx=poly->vtex[lcnt].xy[0]; newy=poly->vtex[lcnt].xy[1];
    if (oldx==newx) oldx+=fudge;/* prevent divide-by-zero in case of vertical line. */

    /* set line endpoints so that x1<x2. */
    if (oldx>newx){
      x1=newx; y1=newy; x2=oldx; y2=oldy;
    }else{
      x1=oldx; y1=oldy; x2=newx; y2=newy;
    }
    /* check if xloc is bracketed by x-endpoints. */
    if (x1<xloc && xloc<=x2){
      /* get formula for line */
      slope=(y2-y1)/(x2-x1);/* dy/dx */ base=y1-(slope*x1);
      intersectY=(xloc*slope)+base;
/*      if (yloc-(y1+ (xloc-x1) * (x2-x1)/(y2-y1) ) >=0); */
      if (intersectY<=yloc){
        inside=!inside;/* toggle inside-ness-flag. */
      }
    }
    oldx=newx; oldy=newy;
  }/* end for lcnt. */

  /* Test if inside is true. If we crossed an odd number of the
  polygon's walls, then the point is inside the polygon. */
  if (inside){
    return (true);
  }else{
    return (false);
  }
}/* inside_poly */
#endif

void testpoly()
/* **************************************************************
*  Just for debugging.
************************************************************** */
{
  polyT poly;
  CubeT minmax;
  int cnt;

  RegPoly(20,20, 10.0,10,&poly);
  get_poly_limits(&poly,&minmax);
  for (cnt=0;cnt<MaxDims;cnt++){
    printf("dimension:%i, min:%1.3lf, max:%1.3lf \n",cnt,minmax.L[0][cnt],minmax.L[1][cnt]);
  }
  printf("starting inside test.\n");
  if (inside_poly(&poly,29.99999999999999,20)){
    printf("inside!\n");
  }else{
    printf("not inside!\n");
  }
  freepoly(&poly);
}

void get_points_inside_poly(VectT *xyvect, PListT *xydata, polyT *poly, VectT *results)
/**************************************************************
*  Scan a spatially-sorted data set for all the points that are
* inside a 2d polygonal region.
**************************************************************/
{
  CubeT poly_minmax;
  /* low-level pointers and variables for quick dereferencing. */
  intdex *V;  XYZType *P;
  intdex cnt,newcnt,NumPoints,datadex;
  
  /* poly must be closed for this to work. */
  poly->closed=true;
  
  /* Get bounding box of polygon. */
  get_poly_limits(poly,&poly_minmax);
  
  /* Narrow it down to just what's in the box.  */
  TraverseZone(xyvect,xydata,&(poly_minmax),results);

  /* Now pack the array down to only those xy's really inside the polygon. */
  NumPoints=results->NumPoints;
  V=&(results->V[0]);/* quicker dereference. */
  newcnt=0;
  for (cnt=0;cnt<NumPoints;cnt++){
    datadex=V[cnt]; P=&(xydata->P[datadex]);
    if (inside_poly(poly,P->Loc[0],P->Loc[1])){
      V[newcnt]=datadex; newcnt++;
    }
  }
  results->NumPoints=newcnt;
}/* get_points_inside_poly */

/* End Polygon support. */

#if 0

enum ClipState {Outside,Clipped,Inside};

void clipline(
  int *lx1,int *ly1,int *lx2,int *ly2,
  int rx1,int ry1,int rx2,int ry2,
  int *ClipState)
/* **************************************************************
*  Clips lines to edge of box.  Assumes box is sorted.
************************************************************** */
{
  double Slope,Base;
  bool LineClipped=false;

  double Lx1=*lx1,Ly1=*ly1,Lx2=*lx2,Ly2=*ly2;

  *ClipState = Outside;

  /*  X - Constrain.  */
  if (!( ((Lx1<rx1) && (Lx2<rx1)) ||
          ((Lx1>rx2) && (Lx2>rx2)) )){

    /*  Calculate slope, base for Y==MX+B.  */
    if (Lx1==Lx2 ){
      Slope = maxlong; /* Infinity. */
    }else{
      Slope = (Ly2-Ly1)/(Lx2-Lx1);
    }
    Base = Ly1-(Lx1*Slope);

    if ((Lx1<rx1) ){  /*  Take intersection with rx1. */
      Lx1 = rx1;
      Ly1 = (Slope*Lx1)+Base;
      LineClipped = true;
    }else{
      if ((Lx1>rx2) ){  /*  Take intersection with rx2. */
        Lx1 = rx2;
        Ly1 = (Slope*Lx1)+Base;
        LineClipped = true;
      }
    }

    /*  Second endpoint X.  */
    if ((Lx2<rx1) ){  /*  Take intersection with rx1. */
      Lx2 = rx1;
      Ly2 = (Slope*Lx2)+Base;
      LineClipped = true;
    }else{
      if ((Lx2>rx2) ){  /*  Take intersection with rx2. */
        Lx2 = rx2;
        Ly2 = (Slope*Lx2)+Base;
        LineClipped = true;
      }
    }

    /*  Y - Constrain.  */
    if (!( ((Ly1<ry1) && (Ly2<ry1)) ||
            ((Ly1>ry2) && (Ly2>ry2)) )){
      /*  If Ycross out of range, or Ly1 out of range anyway.  */
      if ((Ly1<ry1) ){ /*  Take intersection with ry1. */
        Ly1 = ry1;
        Lx1 = (ry1-Base)/Slope;
        LineClipped = true;
      }else{
        if ((Ly1>ry2) ){ /*  Take intersection with ry2. */
          Ly1 = ry2;
          Lx1 = (ry2-Base)/Slope;
          LineClipped = true;
        }
      }

      /*  Second endpoint Y.  */
      /*  If Ycross out of range, or Ly2 out of range anyway.  */
      if ((Ly2<ry1) ){ /*  Take intersection with ry1. */
        Ly2 = ry1;
        Lx2 = (ry1-Base)/Slope;
        LineClipped = true;
      }else{
        if ((Ly2>ry2) ){ /*  Take intersection with ry2. */
          Ly2 = ry2;
          Lx2 = (ry2-Base)/Slope;
          LineClipped = true;
        }
      }
      if (LineClipped ){
        *ClipState = Clipped;
      }else{
        *ClipState = Inside;
      }
    } /*  End if inside Y-range.  */
  } /*  End if inside X-range.  */

  *lx1=Lx1; *ly1=Ly1; *lx2=Lx2; *ly2=Ly2;
}  /*  End of clipline.  */

#endif
